/*
 *
 *
 *   sgdserver.cc
 *
 *   Simple C++ application class to gateway a serial device into a Unix
 *   socket, to allow multiple clients to connect simultaneously.
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


#include "sgdserver.h"
#include "util.h"
#include "signals.h"

#include <fstream>


// namespaces
using namespace my;
using namespace std;

// use for logging
static const std::string app_name("sgd");


//
//
//   server::iosetup(...)
//
//

//
//   Function Pseudocode:
//
//       for port in ports:
//          if port.client_count < client_max:
//              add port.listener to RX mux
//
//          if port.tty_TX_buffer > 0:
//              add tty to TX mux
//          if always_listen OR port.client_count > 0:
//              add tty to RX mux
//
//          for client in port.clients:
//              if client.tx_buffer.size > 0:
//                  add client to TX mux
//              if port.tty_TX_buffer < tty_TX_max:
//                  add client to RX mux
//

void server::iosetup (my::fdset &rset, my::fdset &wset) {
	plist::iterator pi;
	clist::iterator ci;
	port   *pptr = NULL;
	client *cptr = NULL;

	// zero-out both sets
	rset.zero();
	wset.zero();

	#if 0
	cerr << "IO Setup: " << endl;
	#endif

	// for each PORT...
	for (pi = ports.begin(); pi != ports.end(); ++pi) {
		pptr = *pi;

		// add listener, if limits allow
		if ((pptr->config.client_limit == 0) || (pptr->clients.size() < pptr->config.client_limit))
			rset.set(pptr->listener.getfd());

		// add TTY to RX, if needed
		if (pptr->config.always_read || (pptr->clients.size() > 0))
			rset.set(pptr->device.getfd());

		// add TTY to RX, if needed
		if (pptr->tx_buffer.size() > 0)
			wset.set(pptr->device.getfd());

		#if 0
		if (rset.isset(pptr->listener.getfd()))
			cerr << "Listener RX set." << endl;
		if (rset.isset(pptr->device.getfd()))
			cerr << "TTY RX set." << endl;
		if (rset.isset(pptr->device.getfd()))
			cerr << "TTY TX set." << endl;
		cerr << "TTY TX BUFFER = " << pptr->tx_buffer.size() << " bytes." << endl;
		#endif

		// for each CLIENT ON THIS PORT...
		for (ci = pptr->clients.begin(); ci != pptr->clients.end(); ++ci) {
			cptr = *ci;

			if (cptr->tx_buffer.size() > 0)
				wset.set(cptr->sock_obj.getfd());
			if ((pptr->config.max_tx_queue == 0) || (pptr->tx_buffer.size() < pptr->config.max_tx_queue))
				rset.set(cptr->sock_obj.getfd());

			#if 0
			if (rset.isset(cptr->sock_obj.getfd()))
				cerr << "Client RX set." << endl;
			if (rset.isset(cptr->sock_obj.getfd()))
				cerr << "Client TX set." << endl;
			cerr << "Client TX BUFFER = " << cptr->tx_buffer.size() << " bytes." << endl;
			#endif
		}
	}

	#if 0
	cerr << "IO Setup Complete." << endl;
	#endif
}


/*
 *
 *   server::run()
 *
 */

int server::run(void) {
	my::fdset read_set, write_set;
	ssize_t rtn;
	size_t bytes;

	clist::iterator si;
	bool do_cleanup = false;
	std::string message;

	// used for iteration
	port   *pptr = NULL;
	client *cptr = NULL;


	for (;;) {

		//
		// setup and wait for an event
		//

		iosetup(read_set, write_set);
		rtn = select(&read_set, &write_set, NULL, NULL);

		// deal with errors - FIXME: should be exception-handled
		if (rtn < 0) {
			// if error was hard failure, bail out
			if (errno != EINTR) {
				// FATAL: select() error
				message = "select() returned an error: ";
				message += strerror(errno);
				dolog(message);
				return 1;
			}

			// deal with signals
			if (signal_count(SIGTERM) > 0)
				return 0;
			if (signal_count(SIGINT) > 0)
				return 0;

			// otherwise, try again
			continue;
		}

		// if nothing happened, try again
		if (rtn == 0)
			continue;

		// for each PORT...
		for (plist::iterator pi = ports.begin(); pi != ports.end(); ++pi) {
			pptr = *pi;

			// if a new client is connecting, service it
			if (read_set.isset(pptr->listener.getfd())) {
				unixsocket::address client_addr;
				client *client_obj = 0;

				try {

					//
					// create a new client, and accept a new socket for it
					//

					client_obj = new client;
					pptr->listener.accept(client_obj->sock_obj);

				} catch (const std::exception &e) {
					// log any errors
					if (client_obj) {
						delete client_obj;
						client_obj = NULL;
					}
					message = "Error while accepting inbound socket: ";
					message += e.what();
					dolog(message);
				}

				if (client_obj) {
					// push him onto the list of clients
					pptr->clients.push_back(client_obj);
					// log it
					message = "Accepted new client for device ";
					message += pptr->config.s_device;
					message += ", (";
					message += itoa(count(pptr));
					message += '/';
					message += itoa(count());
					message += ") clients connected (port/total).";
					dolog(message);
				}
			}

			// for each CLIENT...
			for (si = pptr->clients.begin(); si != pptr->clients.end(); ++si) {
				// get the socket object from the list
				cptr = *si;

				// if this client has died for some reason, skip it
				if (cptr->dead)
					continue;

				// if client has data to read
				if (read_set.isset(cptr->sock_obj.getfd())) {
					try {
						// read from socket -> TTY transmit buffer
						if (cptr->sock_obj.get(pptr->tx_buffer) == 0)
							cptr->dead = true;
					} catch (const std::exception &e) {
						message = "Socket error reading from client: ";
						message += e.what();
						message += ", disconnecting.";
						dolog(message, true);
						cptr->dead = true;
					}
				}

				// if client has data to write
				if (write_set.isset(cptr->sock_obj.getfd())) {
					try {
						// write to socket <- Socket transmit buffer
						bytes = cptr->sock_obj.put(cptr->tx_buffer);
						if (bytes > 0)
							cptr->tx_buffer.erase(
									cptr->tx_buffer.begin(),
									cptr->tx_buffer.begin() + bytes);
					} catch (const std::exception &e) {
						message = "Socket error writing to client: ";
						message += e.what();
						message += ", disconnecting.";
						dolog(message, true);
						cptr->dead = true;
					}
				}

				// if the client has disconnected, deal with it
				if (cptr->dead) {
					cptr->dead = true;
					do_cleanup = true;
					cptr->sock_obj.disconnect();
				}
			}


			//
			//
			//   TTY SECTION
			//
			//

			// if the TTY has data to send, send it
			if (write_set.isset(pptr->device.getfd())) {
				try {
					bytes = pptr->device.put(pptr->tx_buffer);
					if (bytes > 0)
						pptr->tx_buffer.erase(
								pptr->tx_buffer.begin(),
								pptr->tx_buffer.begin() + bytes);
				} catch (const std::exception &e) {
					message = "Error reading from TTY device ";
					message += pptr->config.s_device;
					message += ": "; 
					message += e.what();
					message += ", quitting now.";
					dolog(message, true);
					return 1;
				}
			}

			// if the TTY has data waiting, service it
			if (read_set.isset(pptr->device.getfd())) {
				try {
					pptr->rx_buffer.clear();
					pptr->device.get(pptr->rx_buffer);

					// copy the received data into each client TX buffer
					if (pptr->rx_buffer.size()) {
						for (si = pptr->clients.begin(); si != pptr->clients.end(); ++si) {
							cptr = *si;   // get the socket object from the list
							cptr->tx_buffer.insert(
									cptr->tx_buffer.end(),
									pptr->rx_buffer.begin(),
									pptr->rx_buffer.end());
							if (pptr->config.max_tx_queue > 0) {
								if (cptr->tx_buffer.size() > pptr->config.max_tx_queue) {
									cptr->dead = true;
									do_cleanup = true;
								}
							}
						}
					}
				} catch (const std::exception &e) {
					message = "Error reading from TTY device ";
					message += pptr->config.s_device;
					message += ": "; 
					message += e.what();
					message += ", quitting now.";
					dolog(message, true);
					return 1;
				}
			}


			//
			//  CLIENT CLEANUP SECTION
			//
			//  if there are dead client objects, clean them up
			//
			//  at this point, we just need to delete client object from
			//  the list and from memory
			//

			if (do_cleanup) {
				// reset the cleanup flag
				do_cleanup = false;

				for (si = pptr->clients.begin(); si != pptr->clients.end(); ++si) {
					// get the socket object from the list
					cptr = *si;

					if (cptr->dead) {
						// remove the client from the list
						pptr->clients.erase(si);
						// reset the list position pointer
						si = pptr->clients.begin();
						// delete the client object
						delete cptr;

						// talk about it
						message = "Client disconnected from device ";
						message += pptr->config.s_device;
						message += ", (";
						message += itoa(count(pptr));
						message += '/';
						message += itoa(count());
						message += ") clients connected (port/total).";
						dolog(message);
					}
				}
			}
		}
	}
}


/*
 *
 *   server::server(server_config)
 *
 */

server::server(const server_config &cfg) {
	const int flags = my::device::not_ctl_tty | my::device::nonblocking;
	std::string message;

	// copy attributes
	s_daemon  = cfg.o_daemon;
	s_stderr  = cfg.o_stderr;
	s_syslog  = cfg.o_syslog;
	s_pidfile = cfg.o_pidfile;

	// open syslog, if needed
	if (cfg.o_syslog)
		openlog(app_name.c_str(), LOG_PID, LOG_DAEMON);

	// for each port defined...
	for (
			server_config::pclist::const_iterator pi = cfg.s_ports.begin();
			pi != cfg.s_ports.end();
			++pi
		) {

		// reference to the current port config
		const port_config &pc = *pi;

		// start building the new port
		port *sp = new port;
		sp->config = pc;

		try {
			sp->device.open(pc.s_device, flags);
			dolog("Opened device: " + pc.s_device);

		} catch (const std::exception &e) {
			message = "Could not open TTY device ";
			message += pc.s_device;
			message += ": ";
			message += e.what();
			s_stderr = true;
			dolog(message);
			exit(1);
		}

		try {
			sp->device.set_options(pc.s_ios);
			dolog("Configured device: " + pc.s_device);

		} catch (const std::exception &e) {
			message = "Could not set TTY options ";
			message += pc.s_device;
			message += ": ";
			message += e.what();
			s_stderr = true;
			dolog(message);
			exit(1);
		}

		try {
			// if the socket exists, delete it
			remove(pc.s_address.c_str());

			//
			//  bind to the socket
			//

			sp->listener.bind(pc.s_address);
			dolog("Bound to socket: " + pc.s_address);

			//
			// change the mode of the socket
			//

			chmod(pc.s_address.c_str(), pc.s_mode);

			//
			//  change socket ownership
			//

			if (pc.s_owner.size() && pc.s_group.size()) {
				my::auth::userinfo  ui(pc.s_owner);
				my::auth::groupinfo gi(pc.s_group);
				if (ui.valid() && gi.valid())
					chown(pc.s_address.c_str(), ui.uid(), gi.gid());
				else
					dolog("Could not change ownership to " + pc.s_owner + ":" + pc.s_group);
			} else {
				if (pc.s_owner.size()) {
					my::auth::userinfo ui(pc.s_owner);
					if (ui.valid())
						chown(pc.s_address.c_str(), ui.uid(), (unsigned)(-1));
					else
						dolog("Could not change owner to " + pc.s_owner);
				}

				if (pc.s_group.size()) {
					my::auth::groupinfo gi(pc.s_group);
					if (gi.valid())
						chown(pc.s_address.c_str(), (unsigned)(-1), gi.gid());
					else
						dolog("Could not change group to " + pc.s_group);
				}
			}

			dolog("Set permissions on socket: " + pc.s_address);

			//
			// start the socket listening
			//

			sp->listener.listen();

		} catch (const std::exception &e) {
			message = "Bind to address \"";
			message += pc.s_address;
			message += "\" failed: ";
			message += e.what();
			s_stderr = true;
			dolog(message);
			exit(1);
		}

		// add the port to the list
		ports.push_back(sp);
	}

	// daemonize
	if (cfg.o_daemon && ! cfg.o_stderr) {
		// ...then do it
		my::daemon::daemon(true).daemonize(true);

		// talk about it..
		dolog("Become daemon: new PID is " + itoa(getpid()));

		ofstream(s_pidfile.c_str()) << (unsigned)getpid() << endl;

		// these should probably be integrated into 'daemon' class
		close(0);
		close(1);
		close(2);
	}
}


/*
 *
 *   server::~server()
 *
 */

server::~server(void) throw() {
	std::string sockfile;

	// remove all sockets and PID files
	for (plist::iterator pi = ports.begin(); pi != ports.end(); ++pi) {
		// remove the device socket
		sockfile = (*pi)->config.s_address;
		dolog("Removing: " + sockfile);
		remove(sockfile.c_str());
	}

	// remove the PID file
	remove(s_pidfile.c_str());
}


/*
 *
 *    server::find(who)
 *
 */

server::client *server::find(const my::unixsocket *who) {
	for (plist::iterator pi = ports.begin(); pi != ports.end(); ++pi) {
		port *p = *pi;
		for (clist::iterator ci = p->clients.begin(); ci != p->clients.end(); ++ci) {
			if ((&((*ci)->sock_obj)) == who)
				return *ci;
		}
	}

	return 0;
}


/*
 *
 *   server::count(who = 0)
 *
 */

size_t server::count(const server::port *who) {
	if (who)
		return (who->clients.size());

	size_t result = 0;

	for (plist::iterator pi = ports.begin(); pi != ports.end(); ++pi) {
		port *p = *pi;
		result += p->clients.size();
	}

	return result;
}


/*
 *
 *   server::dolog(...)
 *
 */

void server::dolog (const std::string &message, bool is_error) {
	if (s_stderr)
		cerr << message << endl;
	if (s_syslog) {
		if (is_error)
			syslog(LOG_ERR, message.c_str());
		else
			syslog(LOG_NOTICE, message.c_str());
	}
}


/*
 *
 *   server::dolog(serialios&)
 *
 */

void server::dolog (/*const*/ my::serialios &ios) {
	char ch;
	std::string message;
	dolog("Speed   : " + itoa(ios.getSpeed()));
	ios.getParity(ch);
	message = "Parity  : ";
	message += ch;
	dolog(message);
	dolog("Wordlen : " + itoa(ios.getCharSize()));
	dolog("Stopbits: " + itoa(ios.getStopBits()));
	message = "SoftFlow: ";
	message += (ios.getSWflow() ? "Yes" : "No");
	dolog(message);
	message = "HardFlow: ";
	message += (ios.getHWflow() ? "Yes" : "No");
	dolog(message);
	message = "Local   : ";
	message += (ios.getLocal() ? "Yes" : "No");
	dolog(message);

	if (s_stderr)
		cerr << ios << endl;
}


// EOF
