/*
 *
 *
 *   sgnetd.cc
 *
 *   PAM helper app for authenticating users to networked SGD sessions.
 *
 *   Copyright (C) 2003, by the Grand River Dam Authority,
 *   All rights reserved.
 *
 *   Author: Matt Roberts
 *
 *
 */


#include <sstream>
#include <iostream>
#include <string>

#include <my++/auth.h>
#include <my++/stype.h>

#include <sys/types.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <grp.h>

#include "util.h"


// application name used by PAM
static const std::string appname("sgnetd");


using namespace my;
using namespace std;

/*
 *
 *   class auth_session
 *
 */

class auth_session : public auth::pam::base_session {
	private:
		std::string hs_app_name, hs_user_name, hs_user_pass;

	public:
		bool doAuth(void);
		auth_session(
				const std::string &app_name,
				const std::string &user_name,
				const std::string &user_pass) :
			base_session(app_name),
			hs_app_name(app_name),
			hs_user_name(user_name),
			hs_user_pass(user_pass) { /* noop */ };
		~auth_session(void) throw() { /* noop */ };

	protected:
		bool conversation (auth::pam::base_session::message_list &ml);
};


bool auth_session::doAuth(void) {
	try {
		// start the authentication sequence
		if (!start(hs_user_name)) return false;

		// set some parameters for the conversation
		//set_user_prompt ("Username: ");

		// authenticate the user
		if (!authenticate(silent)) return false;

		// do account management if needed
		if (!manage_account()) {
			// [ DON'T ] change the user's auth token if needed
			// [ if (!change_token()) ]
				return false;
		}

		// open the user's session
		if (!open_session()) return false;
	} catch (...) {
		try {
			// end the PAM session
			end();
		} catch (...) {
		}

		/* rethrow the exception */
		throw;
	}

	return true;
}


bool auth_session::conversation(auth::pam::base_session::message_list &ml) {
	for (message_list::iterator mi = ml.begin(); mi != ml.end(); ++mi) {
		switch (mi->style) {
			case message::prompt_echo:
				mi->setResponse(hs_user_name);
				break;
			case message::prompt_noecho:
				mi->setResponse(hs_user_pass);
				break;
			default:
				mi->setResponse(0);
				break;
		}
	}

	return true;
}


/*
 *
 *   parse_command(...)
 *
 */

static bool parse_command (
		const std::string &line,
		std::string &cmd,
		std::string &arg) {
	// clear out the return values
	cmd.erase();
	arg.erase();

	// see if we have anything to work with
	if (!line.size())
		return false;

	// local vars
	std::string::iterator si;

	// first, trim the command line
	std::string newline = my::ltrim(my::rtrim(line));

	// see if we have anything left to work with
	if (!line.size())
		return false;

	// find the first space, if there is one
	for (si = newline.begin(); si != newline.end(); ++si) {
		if (isspace(*si))
			break;
	}

	// extract the command
	cmd = my::toLower(std::string(newline.begin(), si));

	// extract the arg, if given
	if (si != newline.end())
		arg = my::ltrim(my::rtrim(std::string(++si, newline.end())));

	// bye!
	return true;
}


/*
 *
 *   parse_open(string, string, bool, bool)
 *
 */

static bool parse_open(const std::string &arg, std::string &sockpath, bool &read_only, bool &persistent) {
	// default values
	read_only = true;
	persistent = false;
	sockpath.erase();

	std::istringstream tokens(arg);
	std::string token;

	tokens >> token;

	if (token.empty())
		return false;
	else
		sockpath = token;


	while (tokens) {
		token.erase();
		tokens >> token;
		token = my::toLower(token);
		if (token == "ro")
			read_only = true;
		else if (token == "rw")
			read_only = false;
		else if (token == "persistent")
			persistent = true;
		else
			return false;
	}

	return true;
}


/*
 *
 *    client_command(argc, argv)
 *

static const std::string client_command (int argc, const char * const * const argv) {
	std::string result;
	int i;

	for (i = 1; i < argc; ++i)
		if (i > 1)
			result += ' ';
		result += argv[i];

	return result;
}
 */


/*
 *
 *    int main (int argc, char **argv)
 *
 */

int main (int argc, char **argv) {
	// open syslog
	openlog(path_file(argv[0]).c_str(), LOG_PID, LOG_USER);

	if (argc < 2) {
		syslog(LOG_ERR, "No child program specified on command line");
		return 1;
	}

	std::string line;
	std::string command, argument;
	std::string user, pass;
	int auth_fail_delay = 2;
	int max_failures = 3;
	int i, j;

	cout << "+OK Hello.\r\n";

	for (i = 0; i < max_failures; ++i) {
		line.erase();
		user.erase();
		pass.erase();

		for (j = 0; j < max_failures; ++j) {
			if (!getline(cin, line))
				return 0;

			if (!parse_command(line, command, argument)) {
				cout << "-ERR Invalid command\r\n";
				continue;
			}

			if (command == "user") {
				if (!argument.size()) {
					cout << "-ERR Command requires an argument\r\n";
					continue;
				}
				cout << "+OK Password please.\r\n";
				user = argument;
				break;
			} else if (command == "quit") {
				cout << "+OK Bye!\r\n";
				return 0;
			} else {
				cout << "-ERR Invalid command\r\n";
				continue;
			}
		}

		if (j == max_failures)
			return 0;

		for (j = 0; j < max_failures; ++j) {
			if (!getline(cin, line))
				return 0;

			if (!parse_command(line, command, argument)) {
				cout << "-ERR Invalid command\r\n";
				continue;
			}

			if (command == "pass") {
				pass = argument;
				break;
			} else if (command == "quit") {
				cout << "+OK Bye!\r\n";
				return 0;
			} else if (command == "user") {
				if (!argument.size()) {
					cout << "-ERR Command requires an argument\r\n";
					continue;
				}
				cout << "+OK We\'ll try that one, too.\r\n";
				j = -1;
				continue;
			} else {
				cout << "-ERR Invalid command\r\n";
				continue;
			}
		}

		if (j == max_failures)
			return 0;

		//cerr << "Authenticating: \'" << appname << "\',\'" << user << "\',\'" << pass << '\'' << endl;

		/* create a PAM session */
		auth_session a (appname, user, pass);
	
		/* check the authenticity of the user */
		try {
			if (!a.doAuth()) {
				// complain to syslog
				syslog(LOG_WARNING, "Login failed for user = %s", user.c_str());
				// wait
				if (auth_fail_delay)
					sleep(auth_fail_delay);
				// display error message
				cout << "-ERR Login failed\r\n";
				// try again
				continue;
			} else
				break;
		} catch (const exception &e) {
			syslog(LOG_ERR, "Error occurred while authenticating user %s: %s", user.c_str(), e.what());
			cout << "-ERR Internal: " << e.what() << "\r\n";
			return 1;
		} catch (...) {
			syslog(LOG_ERR, "Undefined error occurred while authenticating user %s.", user.c_str());
			cout << "-ERR Internal: Undefined error\r\n";
			return 1;
		}
	}

	if (i == max_failures)
		return 0;

	// log our login results
	syslog(LOG_INFO, "Successful logon: user = %s", user.c_str());

	// get user info
	my::auth::userinfo ui(user);

	if (!ui.valid()) {
		syslog(LOG_WARNING, "Can't get UID for user %s.", user.c_str());
		cout << "Can't get UID for user " << user << "\r\n";
		return 1;
	}

	// become user
	syslog(LOG_WARNING, "Become UID,GID: %d,%d.", ui.uid(), ui.gid());
	initgroups(user.c_str(), ui.gid());
	setuid(ui.uid());

	// display next prompt
	cout << "+OK Please send OPEN command.\r\n";


	//
	//
	//   wait for an OPEN command from the user
	//
	//

	std::string sockpath;
	bool read_only = true, persistent = false;

	for (;;) {
		if (!getline(cin, line))
			return 0;

		if (!parse_command(line, command, argument)) {
			cout << "-ERR Invalid command\r\n";
			continue;
		}

		if (command == "open") {
			if (parse_open(argument, sockpath, read_only, persistent)) {
				cout << "-ERR Error while reading command\r\n";
				continue;
			}

			if (argument.empty()) {
				cout << "-ERR Command requires an argument\r\n";
			} else {
				break;
			}
		} else if (command == "quit") {
			cout << "+OK Bye!\r\n";
			return 0;
		} else {
			cout << "-ERR Invalid command\r\n";
			continue;
		}
	}

	//
	//
	//   start new process here
	//
	//

	// get the base name of the command
	command = path_file(argv[1]);

	char const * args[] = { NULL, NULL, NULL, NULL, NULL, NULL };
	i = 0;

	// first argv is command name
	args[i++] = command.c_str();

	// second argv is '-n' to specify network mode sgcd
	args[i++] = "-n";

	// optional command line args
	if (read_only)
		args[i++] = "-r";
	if (persistent)
		args[i++] = "-p";

	// append socket path to command line
	args[i++] = sockpath.c_str();

	// run the command
	if (execvp (command.c_str(), (char * const*)args) < 0) {
		syslog(LOG_ERR, "Could not exec %s %s: %s", command.c_str(), sockpath.c_str(), strerror(errno));
		return 1;
	}

	// this should never happen
	return 0;
}
